#
# A set of helper functions for various DAC operations
#

#
# All public functions go here
#


#
# Create a server-connection given an instance name
#  
function Get-ServerConnection([switch]$help, [string]$serverName)
{   
    if($help)
    {
    	Write-Host "`n[Usage]: Get-ServerConnection -serverName `"server name`" `n"
    	Write-Host "`nThis function creates the server connection on the given server instance using the default windows authentication." 
        Write-Host "If you need a ServerConnection with different credentials, please use [Microsoft.SqlServer.Management.Common.ServerConnection] instead.`n"
    }
    else
    {
        Trap 
        {  
            PrintException($_.Exception);    
            return;  
        }

        $serverConnection = New-Object Microsoft.SqlServer.Management.Common.ServerConnection($serverName)
        return $serverConnection
    }
}


#
# Load a DACType from a DACPac file
#
function Get-DacType([switch]$help, [string]$dacPackPath)
{
    if($help)
    {
    	Write-Host "`n[Usage]: Get-DacType -dacPackPath `"full_path_of_dacpac`" `n"
    	Write-Host "`nThis function loads the DAC Type from the input dacpac file path.`n" 
    }
    else
    {
        Trap 
        {  
            PrintException($_.Exception);    
            return;  
        }
      
        $dacpac = [System.IO.File]::OpenRead($dacPackPath)
        $dac = [Microsoft.SqlServer.Management.Dac.DacType]::Load($dacpac)
        
        return $dac 
    }
}


#
# Save a DACType to a DACPac file
#
function Save-DacType([switch]$help, 
                      [Microsoft.SqlServer.Management.Dac.DacType]$dacType, 
                      [string]$dacPackPath)
{
    if($help)
    {
    	Write-Host "`n[Usage]: Save-DacType -dacType `"dacType_object`" -dacPackPath `"full_path_of_dacpac`"`n"
    	Write-Host "`nThis function saves the DAC Type as a DAC package file to the input dacpac file path.`n" 
    }
    else
    {
        Trap 
        {  
            PrintException($_.Exception);    
            return;  
        }
      
        $dacpac = [System.IO.File]::Create($dacPackPath)
        [Microsoft.SqlServer.Management.Dac.DacType]::Save($dacType, $dacpac)
        $dacpac.Close()
    }
}

#
# Extracts a DAC (Type) from a database. 
#
function Extract-DAC([switch]$help, 
                     [Microsoft.SqlServer.Management.Common.ServerConnection]$serverConnection, 
                     [string]$databaseName, 
                     [string]$dacTypeName, 
                     [System.Version]$dacTypeVersion,
                     [string]$dacPackPath)
{
    if($help)
    {
    	Write-Host "`n[Usage]: Extract-DacType -serverConnection `"serverConnection`" -databaseName `"databaseName`" -dacTypeName `"dacTypeName`" -dacTypeVersion `"dacTypeVersion`" -dacPackPath `"dacPackPath`"`n"
    	Write-Host "`nThis function extracts a given database on the input connection to a DAC Type and also saves it as a DAC package file to the input dacpac file path. The user can also specify the DAC Type name and version (System.Version).`n" 
    }
    else
    {
        Trap 
        {  
            PrintException($_.Exception);    
            return;  
        }
      
        $server = GetSmoServer($serverConnection)
        $dacExtractionUnit = new-object Microsoft.SqlServer.Management.Dac.DacExtractionUnit($server, $databaseName, $dacTypeName, $dacTypeVersion)
        
    	#subscribe to the validation event
        Write-Host `n
        $dacExtractionUnit.add_DatabaseObjectValidated({Write-Host ($_.ValidatedObject.Name) is validated with a Result of $_.Result})

        $dacExtractionResult = $dacExtractionUnit.Extract($dacPackPath)
      
        $dacType = Get-DacType($dacPackPath)
        Write-Host `n`n***Succesfully Extracted Database***`n
        return $dacType
    }
}

#
# Registers an existing database as a DAC instance. 
# Returns the DAC Type
# 
function Register-Dac([switch]$help,
                      [Microsoft.SqlServer.Management.Common.ServerConnection]$serverConnection, 
                      [string]$databaseName, 
                      [string]$dacTypeName, 
                      [string]$dacTypeVersion)
{
    if($help)
    {
    	Write-Host "`n[Usage]: Register-Dac -serverConnection `"serverConnection`" -databaseName `"databaseName`" -dacTypeName `"dacTypeName`" -dacTypeVersion `"dacTypeVersion`"`n"
    	Write-Host "`nThis function registers a given database on the input connection as a DAC instance. The user can also specify the DAC type name and version (System.Version) of the new DAC instance.`n" 
    }
    else
    {
        Trap 
        {  
            PrintException($_.Exception);    
            return;  
        }
        
        $server = GetSmoServer($serverConnection)
        $dacExtractionUnit = new-object Microsoft.SqlServer.Management.Dac.DacExtractionUnit($server, $databaseName, $dacTypeName, $dacTypeVersion)
        
        #subscribe to the validation event
        Write-Host `n
        $dacExtractionUnit.add_DatabaseObjectValidated({Write-Host ($_.ValidatedObject.Name) is validated with a Result of $_.Result})

        $dacExtractionResult = $dacExtractionUnit.Register()
      
        $serverConnection.Disconnect()
        $serverConnection.Connect()
        
        $dacStore = new-object Microsoft.SqlServer.Management.Dac.DacStore $serverConnection
        if ($dacStore.DacInstances.Contains($databaseName))
        {
            $dacInstance = $dacStore.DacInstances[$databaseName]
        
            Write-Host `n`n***Succesfully Registered Database***`n
            return $dacInstance.Type
        }
        
        return null
    }
}

#
# Adds an extra file to give DAC type
#
function AddExtraFile-Dac([switch]$help,
                          [Microsoft.SqlServer.Management.Dac.DacType]$dacType, 
                          [string]$filePath,
                          [string]$tag)
{
    if($help)
    {
    	Write-Host "`n[Usage]: AddExtraFile-Dac -dacType `"dacType_object`" -filePath `"extrafile Path`" -tag `"tag for extrafile`"`n"
    	Write-Host "`nThis function adds an extra file to a given DAC type. There is no restriction on the file content. Users can add multiple extra files. However, the size of all files together cannot exceed 50 MB." 
    	Write-Host "`nThe user can also provide a tag for the extra file being added. `n" 
    }
    else
    {
        Trap 
        {  
            PrintException($_.Exception);    
            return;  
        }
        
        $extraSource = New-Object Microsoft.SqlServer.Management.Dac.FileExtraSource($filePath, [System.IO.Path]::GetFileName($filePath), $tag)
        
        $filesCollection = New-Object Microsoft.SqlServer.Management.Dac.MiscellaneousFilesCollection
        
        foreach($file in $dacType.MiscellaneousFiles)
        {
            $filesCollection.Add($file)
        }
        
        $filesCollection.Add($extraSource)
        
        $newDacType = New-Object Microsoft.SqlServer.Management.Dac.DacType($dacType.Definition, $dacType.Name,
                                 $dacType.Version, $dacType.Description, $dacType.TargetSelectionPolicy, $filesCollection);
            
        return $newDacType;    
    }
}

#
# Adds post-deploy script file to give DAC type
#
function AddPostDeployFile-Dac([switch]$help,
                          [Microsoft.SqlServer.Management.Dac.DacType]$dacType, 
                          [string]$filePath)
{
    if($help)
    {
    	Write-Host "`n[Usage]: AddPostDeployFile-Dac -dacType `"dacType_object`" -filePath `"extrafile Path`"`n"
    	Write-Host "`nThis function is a special case of AddExtraFile-Dac. The function adds an extra file to a given DAC type with explicit `"POST-DEPLOY`" tag." 
        Write-Host "`nWhile deploying (See Deploy-Dac) a DAC Type, any file with this tag is run against the DAC database after the DAC schema is deployed. Hence, the file should be a Transact-SQL script.`n" 
    }
    else
    {
        Trap 
        {  
            PrintException($_.Exception);    
            return;  
        }
        
        $extraSource = New-Object Microsoft.SqlServer.Management.Dac.FileExtraSource($filePath, [System.IO.Path]::GetFileName($filePath), "POST-DEPLOY")
        
        $filesCollection = New-Object Microsoft.SqlServer.Management.Dac.MiscellaneousFilesCollection
        
        foreach($file in $dacType.MiscellaneousFiles)
        {
            $filesCollection.Add($file)
        }
        
        $filesCollection.Add($extraSource)
        
        $newDacType = New-Object Microsoft.SqlServer.Management.Dac.DacType($dacType.Definition, $dacType.Name,
                                 $dacType.Version, $dacType.Description, $dacType.TargetSelectionPolicy, $filesCollection);
            
        return $newDacType;    
    }        
}
                   

#
# Deploys a DACType to the server. 
#  If the DAC instance already exists on the server, the DAC instance is upgraded. 
#  Otherwise, a brand new DAC instance is created on the server
#
function Deploy-Dac([switch]$help,
                   [Microsoft.SqlServer.Management.Common.ServerConnection]$serverConnection, 
                   [Microsoft.SqlServer.Management.Dac.DacType]$dacType, 
                   [string]$dacInstanceName)
{

    if($help)
    {
    	Write-Host "`n[Usage]: Deploy-Dac -serverConnection `"serverConnection`" -dacType `"dacType_object`" -dacInstanceName `"dacInstanceName`"`n"
    	Write-Host "`nThis function deploys a given DAC Type on the connection provided."
        Write-Host "`nIf a DAC instance with the name dacInstanceName does not already exists, the DACType is installed. Otherwise, the DACType is used to upgrade the existing DAC instance. However, in order to upgrade the existing DAC instance, the DAC Type name must match the name of the DAC Type already installed on the instance.`n" 
    }
    else
    {
        Trap 
        {  
            PrintException($_.Exception);    
            return;  
        }
        
        $dacStore = new-object Microsoft.SqlServer.Management.Dac.DacStore $serverConnection  

    	#subscribe
        Write-Host `n
    	$dacStore.add_DacActionStarted({Write-Host `nStarting at $(get-date) :: $_.Description})
    	$dacStore.add_DacActionFinished({Write-Host Completed at $(get-date) :: $_.Description})

        if ($dacStore.DacInstances.Contains($dacInstanceName))
        {
            $dacUpgradeOptions = New-Object Microsoft.SqlServer.Management.Dac.DacUpgradeOptions
            $dacInstance = UpgradeDacInternal $dacStore $dacType $dacInstanceName $dacUpgradeOptions
        }
        else
        {
            $dacInstance = InstallDacInternal $dacStore $dacType $dacInstanceName 
        }
          
        
        ExecutePostDeployFile $serverConnection $dacInstanceName $dacType
          
        Write-Host `n***Succesfully Deployed DAC***`n
        return $dacInstance
    }        
}

#
# Uninstalls an existing DAC 
#   The underlying database is dropped/detached depending on the uninstall mode
#
function Uninstall-Dac([switch]$help,
                       [Microsoft.SqlServer.Management.Common.ServerConnection]$serverConnection, 
                       [string]$dacInstanceName,
                       [Microsoft.SqlServer.Management.Dac.DacUninstallMode]$uninstallMode)
{

    if($help)
    {
    	Write-Host "`n[Usage]: Uninstall-Dac -serverConnection `"serverConnection`" -dacInstanceName `"dacInstanceName`" -uninstallMode`"uninstallMode`"`n"
    	Write-Host "`nThis function uninstalls a given DAC instance on the connection provided."
        Write-Host "`nUninstallMode can take three values." 
        Write-Host "`n[Microsoft.SqlServer.Management.Dac.DacUninstallMode]::MakeUnmanaged - This option simply unregisters the DAC without making changes to the database." 
        Write-Host "[Microsoft.SqlServer.Management.Dac.DacUninstallMode]::DetachDatabase - This option simply unregisters the DAC and detaches the database." 
        Write-Host "[Microsoft.SqlServer.Management.Dac.DacUninstallMode]::DropDatabase - This option simply unregisters the DAC and drops the database.`n" 
    }
    else
    {
        Trap 
        {  
            PrintException($_.Exception);    
            return;  
        }
        
        $dacStore = new-object Microsoft.SqlServer.Management.Dac.DacStore $serverConnection
      
    	#subscribe
        Write-Host `n
    	$dacStore.add_DacActionStarted({Write-Host `nStarting at $(get-date) :: $_.Description})
    	$dacStore.add_DacActionFinished({Write-Host Completed at $(get-date) :: $_.Description})
        
        $dacStore.Uninstall($dacInstanceName, $uninstallMode)
        
        Write-Host `n***Succesfully Uninstalled DAC***`n
    }        
}


#
# Unpacks a DAC Type to a given folder
#
function Unpack-Dac([switch]$help,
                    [Microsoft.SqlServer.Management.Dac.DacType]$dacType, 
                    [string]$unpackFolderPath)
{
    if($help)
    {
    	Write-Host "`n[Usage]: Unpack-Dac -dacType `"dacType`" -unpackFolderPath `"unpackFolderPath`"`n"
    	Write-Host "`nThis function unpacks an input DAC package (dacpac) file into a folder by extracting the related metadata xml files and any extra files as well.`n"
    }
    else
    {
        Trap 
        {  
            PrintException($_.Exception);    
            return;  
        }
        
        [Microsoft.SqlServer.Management.Dac.DacType]::Unpack($dacType, $unpackFolderPath);
        
        Write-Host `n***Succesfully Unpacked DAC***`n
    }        
}

#
# Builds a DAC from the input TSQL files.
#
function Build-Dac([switch]$help,
                   [string]$dacTypeName,
                   [array]$files)
{    
    if($help)
    {
    	Write-Host "`n[Usage]: Build-Dac -dacTypeName `"dacTypeName`" -files `"@{file1, file2,...}`"`n"
    	Write-Host "`nThis function compiles the content of Transact-SQL input files and builds a DAC Type. Any build errors are written to the output.`n"
    }
    else
    {
        Trap 
        {  
            PrintException($_.Exception);    
            return;  
        }

        $dacUnit = New-Object Microsoft.SqlServer.Management.Dac.DacCompilationUnit($dacTypeName, 
                        "1.0.0.0",
                        "SQL_Latin1_General_CP1_CI_AS" )
        
        foreach ($file in $files)
        {
            $dacUnit.AddTSqlSourceFile($file, [System.IO.File]::ReadAllText($file)) | out-null
        }
                        
        $dacResult = $dacUnit.Compile()

        if($dacResult.DacType -eq $null)
        {
        	foreach($error in $dacResult.Errors)
        	{
        		$errorInfo = $error.SourceInfo
        		
        		Write-Host "`n"
        		Write-Host("Error found in file : " + $errorInfo.FileName)
        		Write-Host("Start Location : " + $errorInfo.Start.ToString())
        		Write-Host("End Location : " + $errorInfo.End.ToString())
        		Write-Host("Error Message : " + $error.Message)
        		Write-Host("IsWarning : " + $error.IsWarning)	
        		Write-Host "`n"

        		return $null
        	}
        }
        
        Write-Host `n***Succesfully Compiled DAC***`n

        return $dacResult.DacType
    }        
}

#
# All internal functions go here
#

function InstallDacInternal([Microsoft.SqlServer.Management.Dac.DacStore]$dacStore,
                   [Microsoft.SqlServer.Management.Dac.DacType]$dacType, 
                   [string]$dbname)
{
  $deploymentProperties = new-object Microsoft.SqlServer.Management.Dac.DatabaseDeploymentProperties($dacStore.ServerConnection, $dbname)
  $dacInstance = $dacStore.Install($dacType, $deploymentProperties, $false)

  return $dacInstance
}

function UpgradeDacInternal([Microsoft.SqlServer.Management.Dac.DacStore]$dacStore, 
                    [Microsoft.SqlServer.Management.Dac.DacType]$dacType, 
                    [string]$dacInstanceName,
                    [Microsoft.SqlServer.Management.Dac.DacUpgradeOptions]$dacUpgradeOptions)
{
   $upgradeResult = $dacStore.Upgrade($dacInstanceName, $dacType, $dacUpgradeOptions)
 
   # Need to drop the old database now
   $server = new-object Microsoft.SqlServer.Management.Smo.Server($dacStore.ServerConnection)
   $server.KillDatabase($upgradeResult.OriginalDatabaseName)

   $dacInstance = $upgradeResult.UpgradedDacInstance
   return $dacInstance
}

#
# Returns a SMO.Server for a given ServerConnection
#
function GetSmoServer([Microsoft.SqlServer.Management.Common.ServerConnection]$serverConnection)
{
  if($serverConnection -eq $null) {throw New-Object System.ArgumentNullException("serverConnection");}
  
  $serverConnection.Connect();
  $server = new-object Microsoft.SqlServer.Management.Smo.Server($serverConnection)
  return $server
}

#
# Execute the POST-DEPLOY script
#
function ExecutePostDeployFile([Microsoft.SqlServer.Management.Common.ServerConnection]$serverConnection, 
                     [string]$databaseName, 
                     [Microsoft.SqlServer.Management.Dac.DacType]$dacType)
{
    [void]$serverConnection.Connect();
    [void]$serverConnection.SqlConnectionObject.ChangeDatabase($databaseName);

    foreach($miscFile in $dacType.MiscellaneousFiles)
    {
        if($miscFile.Tags.Contains("POST-DEPLOY"))
        {
            $content = [System.Text.Encoding]::UTF8.GetString($miscFile.GetContents());            
            [void]$serverConnection.ExecuteNonQuery($content);                
        }
    }        
}

#
# Prints Complete Stack of a given exception
#
function PrintException([Exception]$exception)
{
    if($exception -eq $null) { return; }
    
    write-output $exception.Message`n
    while ( $exception.InnerException )    
    {    
        $exception = $exception.InnerException    
        write-output $exception.Message`n
    };           
}


