###########################################################
# SQL 2008 Powershell Module for Policy Based Management 
# Bob Beauchemin 
#
###########################################################

######################################################
# Enumerated constants
######################################################
$optype_eq = [Microsoft.SqlServer.Management.Dmf.OperatorType]::EQ
$optype_and = [Microsoft.SqlServer.Management.Dmf.OperatorType]::AND

$enum_false = [Microsoft.SqlServer.Management.Dmf.ExpressionNodeFunction+Function]::False
$func_false = new-object ('Microsoft.SqlServer.Management.Dmf.ExpressionNodeFunction') $enum_false

######################################################
# Gets a policy store object for a specific server
# Currently integrated security only
######################################################
function Get-PolicyStore
{
        Param($serverName)

        $server = New-Object Microsoft.SqlServer.Management.Smo.Server($serverName);
        $server.ConnectionContext.LoginSecure = $true;
        $server.ConnectionContext.Connect();
        
        $sqlStoreConnection = New-Object Microsoft.SqlServer.Management.Sdk.Sfc.SqlStoreConnection($server.ConnectionContext.SqlConnectionObject);
        
        $policyStore = New-Object Microsoft.SqlServer.Management.Dmf.PolicyStore($sqlStoreConnection)
 
        $policyStore;
}

######################################################
# Gets a disconnected policy store object 
######################################################
function Get-OfflinePolicyStore
{
        $policyStore = new-object ('Microsoft.SqlServer.Management.Dmf.PolicyStore')
        ([Microsoft.SqlServer.Management.Sdk.Sfc.SfcConnectionContext]$policyStore).Mode = [Microsoft.SqlServer.Management.Sdk.Sfc.SfcConnectionContextMode]::Offline
 
        $policyStore;
}

######################################################
# Creates a new condition based on Specified Facet
# Online - Creates the condition with Create rather than adding it to the collection
######################################################
function New-Condition
{
        Param($policyStore, $conditionName, $facetName, $expNode)
        
        $condition = new-object ('Microsoft.SqlServer.Management.Dmf.Condition') $policyStore, $conditionName
        $condition.Facet =  $facetName
        $condition.ExpressionNode = $expNode
        $condition.Create()
        
        $condition;
}

######################################################
# Creates a new condition based on Specified Facet
# Adds it to the collection
######################################################
function New-ConditionOffline
{
        Param($policyStore, $conditionName, $facetName, $expNode)
        
        $condition = new-object ('Microsoft.SqlServer.Management.Dmf.Condition') $policyStore, $conditionName
        $condition.Facet =  $facetName
        $condition.ExpressionNode = $expNode
        $policyStore.Conditions.Add($condition)
        
        $condition;
}

######################################################
# Gets corresponding facet name for the facet that is referred in the condition
######################################################
function Get-FacetName
{
        Param($condition)
        
        $facetInfo = [Microsoft.SqlServer.Management.Dmf.PolicyStore]::Facets[$condition.Facet];
        
        $facetInfo.FacetType.Name;
}

######################################################
# Creates a new objectset based on the facet that is referred by the condition
######################################################
function New-ObjectSet
{
        Param($policyStore, $policyName, $condition)
        
        $objectSetName = $policyName + "_ObjectSet"
        $objectSet = new-object ('Microsoft.SqlServer.Management.Dmf.ObjectSet')($policyStore, $objectSetName)
        $objectSet.Facet = GetFacetName $condition
        $policyStore.ObjectSets.Add($objectSet)
        
        $objectSet;
}

######################################################
# Creates a new policy object and sets the condition and object set references
######################################################
function New-Policy
{
        Param($policyStore, $policyName, $condition, $objectSet)
        
        $policy = new-object ('Microsoft.SqlServer.Management.Dmf.Policy') $policyStore, $policyName
        $policy.Condition = $condition
        if ($objectSet -ne $null)
            {
            $policy.ObjectSet = $objectSet.Name
            }
        $policy.AutomatedPolicyEvaluationMode = [Microsoft.SqlServer.Management.Dmf.AutomatedPolicyEvaluationMode]::None
        $policy.Enabled = $false    
        $policy.Create()

        $policy;
}

######################################################
# Creates a new policy object and sets the condition and object set references
######################################################
function New-PolicyOffline
{
        Param($policyStore,  $condition, $objectSet)
        
        $policy = new-object ('Microsoft.SqlServer.Management.Dmf.Policy')($policyStore, $policyName)
        $policy.Condition = $condition.Name
        $policy.ObjectSet = $objectSet.Name
        $policyStore.Policies.Add($policy)

        $policy;
}

######################################################
# Creates a new policy category
######################################################
function New-PolicyCategory
{
        Param($policyStore,  $categoryName)
        
        $cat = new-object ('Microsoft.SqlServer.Management.Dmf.PolicyCategory')($policyStore, $categoryName)
	$cat.MandateDatabaseSubscriptions = $false
	$cat.Create()

        $cat;
}

######################################################
# Saves given policy object to a file. 
# To view this policy file, you can launch SQL Server 2008 Management Studio, 
#       and click on menu File -> Open -> Policy, Browse to %TEMP% to locate this xml file
######################################################
function Export-PolicyToFile
{
        Param($policy,  $policyFilePath)
        
        $xmlWriter = [System.Xml.XmlWriter]::Create($policyFilePath)
        $policy.Serialize($xmlWriter)
        $xmlWriter.Close()
        write-host "Policy Saved to file:" + $policyFilePath
}






